package core.engine.toolbox;

import core.engine.Engine;
import core.engine.entity.Camera;
import org.lwjgl.system.windows.MOUSEINPUT;
import utils.MatrixGraphicUtils;
import utils.vectors.Matrix4f;
import utils.vectors.Vector2f;
import utils.vectors.Vector3f;
import utils.vectors.Vector4f;

public class MousePicker {

    private Vector3f currentRay;
    private Matrix4f projectionMatrix;
    private Matrix4f viewMatrix;
    private Camera camera;

    public MousePicker(Camera camera, Matrix4f projectionMatrix) {
        this.camera = camera;
        this.projectionMatrix = projectionMatrix;
        this.viewMatrix = MatrixGraphicUtils.createViewMatrix(camera);
    }

    public Vector3f getCurrentRay() {
        return currentRay;
    }

    public void update(float mouseX, float mouseY) {
        this.viewMatrix = MatrixGraphicUtils.createViewMatrix(camera);
        this.currentRay = calculateMouseRay(mouseX, mouseY);
    }

    public Vector3f calculateMouseRay(float mouseX, float mouseY) {
        Vector2f normalizedDeviceCoords = getNormalizedDeviceCoords(mouseX, mouseY);
        Vector4f clipCoords = new Vector4f(normalizedDeviceCoords.x, normalizedDeviceCoords.y, -1f, 1f);
        Vector4f eyeCoords = toEyeCoords(clipCoords);
        return toWorldCoords(eyeCoords);
    }

    private Vector2f getNormalizedDeviceCoords(float mouseX, float mouseY) {
        float x = (2f* mouseX) / Engine.WINDOW_WIDTH -1;
        float y = (2f * mouseY) / Engine.WINDOW_HEIGHT -1;
        return new Vector2f(x, y);
    }

    private Vector4f toEyeCoords(Vector4f clipCoords) {
        Matrix4f invertedProjectionMatrix = Matrix4f.invert(projectionMatrix, null);
        Vector4f eyeCoords = Matrix4f.transform(invertedProjectionMatrix, clipCoords, null);
        return new Vector4f(eyeCoords.x, eyeCoords.y, -1f, 0f);
    }

    private Vector3f toWorldCoords (Vector4f eyeCoords) {
        Matrix4f invertedViewMatrix = Matrix4f.invert(viewMatrix, null);
        Vector4f rayWorld = Matrix4f.transform(invertedViewMatrix, eyeCoords, null);
        Vector3f worldCoords =  new Vector3f(rayWorld.x, rayWorld.y, rayWorld.z);
        worldCoords.normalise();
        return worldCoords;
    }
}
